@@ -53,6 +53,13 @@ def get_model_by_path(schema: Dict[str, Any], keys: List[str]) -> Dict[str, Any]
     return get_model_by_path(schema[keys[0]], keys[1:])
 
 
+SPECIAL_PATH_FORMAT: str = '#-datamodel-code-generator-#-{}-#-special-#'
+
+
+def get_special_path(keyword: str, path: List[str]) -> List[str]:
+    return [*path, SPECIAL_PATH_FORMAT.format(keyword)]
+
+
 json_schema_data_formats: Dict[str, Dict[str, Types]] = {
     'integer': {
         'int32': Types.int32,
@@ -367,46 +374,15 @@ class JsonSchemaParser(Parser):
         if obj.title:
             self.extra_template_data[name]['title'] = obj.title
 
-    def parse_list_item(
-        self, name: str, target_items: List[JsonSchemaObject], path: List[str]
-    ) -> DataType:
-        def parse_item(index: int, item: JsonSchemaObject) -> DataType:
-            if item.ref:  # $ref
-                return self.get_ref_data_type(item.ref)
-            elif not any(v for k, v in vars(item).items() if k != 'type'):
-                # trivial types
-                return self.get_data_type(item)
-            elif (
-                item.is_array
-                and isinstance(item.items, JsonSchemaObject)
-                and not any(v for k, v in vars(item.items).items() if k != 'type')
-            ):
-                # trivial item types
-                return self.data_type(
-                    data_types=[self.get_data_type(item.items)], is_list=True,
-                )
-            elif item.is_object:
-                return self.parse_object(
-                    name, item, [*path, str(index)], singular_name=True,
-                )
-            else:
-                return self.get_data_type(item)
-
-        return self.data_type(
-            data_types=[
-                parse_item(index, item) for index, item in enumerate(target_items)
-            ]
-        )
-
     def parse_any_of(
         self, name: str, obj: JsonSchemaObject, path: List[str]
-    ) -> DataType:
-        return self.parse_list_item(name, obj.anyOf, path)
+    ) -> List[DataType]:
+        return self.parse_list_item(name, obj.anyOf, path, obj)
 
     def parse_one_of(
         self, name: str, obj: JsonSchemaObject, path: List[str]
-    ) -> DataType:
-        return self.parse_list_item(name, obj.oneOf, path)
+    ) -> List[DataType]:
+        return self.parse_list_item(name, obj.oneOf, path, obj)
 
     def parse_all_of(
         self,
@@ -417,7 +393,7 @@ class JsonSchemaParser(Parser):
     ) -> DataType:
         fields: List[DataModelFieldBase] = []
         base_classes: List[Reference] = []
-        if len(obj.allOf) == 1:
+        if len(obj.allOf) == 1 and not obj.properties:
             single_obj = obj.allOf[0]
             if single_obj.ref and single_obj.ref_type == JSONReference.LOCAL:
                 if get_model_by_path(self.raw_obj, single_obj.ref[2:].split('/')).get(
@@ -433,6 +409,10 @@ class JsonSchemaParser(Parser):
                         all_of_item, path, get_module_name(name, None),
                     )
                 )
+        if obj.properties:
+            fields.extend(
+                self.parse_object_fields(obj, path, get_module_name(name, None))
+            )
         # ignore an undetected object
         if ignore_duplicate_model and not fields and len(base_classes) == 1:
             return self.data_type(reference=base_classes[0])
@@ -464,127 +444,30 @@ class JsonSchemaParser(Parser):
         exclude_field_names: Set[str] = set()
         for original_field_name, field in properties.items():
 
-            constraints: Optional[Mapping[str, Any]] = None
+            if field.is_array or (
+                self.field_constraints
+                and not (
+                    field.ref
+                    or field.anyOf
+                    or field.oneOf
+                    or field.allOf
+                    or field.is_object
+                    or field.enum
+                )
+            ):
+                constraints: Optional[Mapping[str, Any]] = field.dict()
+            else:
+                constraints = None
+
             field_name, alias = self.model_resolver.get_valid_field_name_and_alias(
                 original_field_name, exclude_field_names
             )
             modular_name = f'{module_name}.{field_name}' if module_name else field_name
 
             exclude_field_names.add(field_name)
-            if field.ref:
-                field_type = self.get_ref_data_type(field.ref)
-            elif field.is_array:
-                field_type = self.parse_array_fields(
-                    modular_name, field, [*path, field_name]
-                ).data_type
-                constraints = field.dict()
-            elif field.anyOf:
-                field_type = self.parse_any_of(modular_name, field, [*path, field_name])
-            elif field.oneOf:
-                field_type = self.parse_one_of(modular_name, field, [*path, field_name])
-            elif field.allOf:
-                field_type = self.parse_all_of(
-                    modular_name,
-                    field,
-                    [*path, field_name],
-                    ignore_duplicate_model=True,
-                )
-            elif field.is_object:
-                if field.properties:
-                    field_type = self.parse_object(
-                        modular_name, field, [*path, field_name]
-                    )
-                elif isinstance(field.additionalProperties, JsonSchemaObject):
-                    field_class_name = self.model_resolver.add(
-                        [*path, field_name], modular_name, class_name=True
-                    ).name
 
-                    # TODO: supports other type
-                    if (
-                        isinstance(field.additionalProperties.items, JsonSchemaObject)
-                        and field.additionalProperties.items.ref
-                    ):
-                        additional_properties_type = self.data_type(
-                            reference=self.model_resolver.add_ref(
-                                field.additionalProperties.items.ref,
-                            ),
-                            is_list=True,
-                        )
-                    elif field.additionalProperties.is_array:
-                        additional_properties_type = self.parse_array(
-                            field_class_name,
-                            field.additionalProperties,
-                            [*path, field_name],
-                        )
-                    elif field.additionalProperties.is_object:
-                        additional_properties_type = self.parse_object(
-                            field_class_name,
-                            field.additionalProperties,
-                            [*path, field_name],
-                            additional_properties=None
-                            if field.additionalProperties.ref
-                            else field,
-                        )
-                    elif field.additionalProperties.enum:
-                        if self.should_parse_enum_as_literal(
-                            field.additionalProperties
-                        ):
-                            additional_properties_type = self.data_type(
-                                literals=field.additionalProperties.enum,
-                            )
-                        else:
-                            additional_properties_type = self.parse_enum(
-                                field_class_name,
-                                field.additionalProperties,
-                                [*path, field_name],
-                            )
-                    elif field.additionalProperties.anyOf:
-                        additional_properties_type = self.parse_any_of(
-                            field_class_name,
-                            field.additionalProperties,
-                            [*path, field_name],
-                        )
-                    elif field.additionalProperties.oneOf:
-                        additional_properties_type = self.parse_one_of(
-                            field_class_name,
-                            field.additionalProperties,
-                            [*path, field_name],
-                        )
-                    elif field.additionalProperties.allOf:
-                        additional_properties_type = self.parse_all_of(
-                            field_class_name,
-                            field.additionalProperties,
-                            [*path, field_name],
-                            ignore_duplicate_model=True,
-                        )
-                    else:
-                        additional_properties_type = self.parse_root_type(
-                            field_class_name,
-                            field.additionalProperties,
-                            [*path, field_name],
-                            additional_properties=None
-                            if field.additionalProperties.ref
-                            else field,
-                        )
-                    self.parse_ref(
-                        field.additionalProperties, [*path, field_name],
-                    )
-                    field_type = self.data_type(
-                        data_types=[additional_properties_type], is_dict=True
-                    )
-                else:
-                    field_type = self.data_type_manager.get_data_type(Types.object)
-            elif field.enum:
-                if self.should_parse_enum_as_literal(field):
-                    field_type = self.data_type(literals=field.enum)
-                else:
-                    field_type = self.parse_enum(
-                        modular_name, field, [*path, field_name]
-                    )
-            else:
-                field_type = self.get_data_type(field)
-                if self.field_constraints:
-                    constraints = field.dict()
+            field_type = self.parse_item(modular_name, field, [*path, field_name])
+
             if self.force_optional_for_required_fields or (
                 self.apply_default_values_for_required_fields and field.has_default
             ):
@@ -618,7 +501,6 @@ class JsonSchemaParser(Parser):
         path: List[str],
         singular_name: bool = False,
         unique: bool = True,
-        additional_properties: Optional[JsonSchemaObject] = None,
     ) -> DataType:
         if not unique:  # pragma: no cover
             warn(
@@ -631,7 +513,7 @@ class JsonSchemaParser(Parser):
         )
         class_name = reference.name
         self.set_title(class_name, obj)
-        self.set_additional_properties(class_name, additional_properties or obj)
+        self.set_additional_properties(class_name, obj)
         data_model_type = self.data_model_type(
             reference=reference,
             fields=self.parse_object_fields(
@@ -646,48 +528,94 @@ class JsonSchemaParser(Parser):
         self.results.append(data_model_type)
         return self.data_type(reference=reference)
 
+    def parse_item(
+        self,
+        name: str,
+        item: JsonSchemaObject,
+        path: List[str],
+        singular_name: bool = False,
+        parent: Optional[JsonSchemaObject] = None,
+    ) -> DataType:
+        if (
+            parent
+            and item.has_constraint
+            and (parent.has_constraint or self.field_constraints)
+        ):
+            root_type_path = get_special_path('array', path)
+            return self.parse_root_type(
+                self.model_resolver.add(
+                    root_type_path, name, class_name=True, singular_name=singular_name,
+                ).name,
+                item,
+                root_type_path,
+            )
+        elif item.ref:
+            return self.get_ref_data_type(item.ref)
+        elif item.is_array:
+            return self.parse_array_fields(
+                name, item, get_special_path('array', path)
+            ).data_type
+        elif item.anyOf:
+            return self.data_type(
+                data_types=self.parse_any_of(
+                    name, item, get_special_path('anyOf', path)
+                )
+            )
+        elif item.oneOf:
+            return self.data_type(
+                data_types=self.parse_one_of(
+                    name, item, get_special_path('oneOf', path)
+                )
+            )
+        elif item.allOf:
+            all_of_path = get_special_path('allOf', path)
+            return self.parse_all_of(
+                self.model_resolver.add(
+                    all_of_path, name, singular_name=singular_name, class_name=True
+                ).name,
+                item,
+                all_of_path,
+                ignore_duplicate_model=True,
+            )
+        elif item.is_object:
+            object_path = get_special_path('object', path)
+            if item.properties:
+                return self.parse_object(
+                    name, item, object_path, singular_name=singular_name
+                )
+            elif isinstance(item.additionalProperties, JsonSchemaObject):
+                return self.data_type(
+                    data_types=[
+                        self.parse_item(name, item.additionalProperties, object_path)
+                    ],
+                    is_dict=True,
+                )
+            return self.data_type_manager.get_data_type(Types.object)
+        elif item.enum:
+            if self.should_parse_enum_as_literal(item):
+                return self.data_type(literals=item.enum)
+            return self.parse_enum(
+                name, item, get_special_path('enum', path), singular_name=singular_name
+            )
+        return self.get_data_type(item)
+
+    def parse_list_item(
+        self,
+        name: str,
+        target_items: List[JsonSchemaObject],
+        path: List[str],
+        parent: JsonSchemaObject,
+    ) -> List[DataType]:
+        return [
+            self.parse_item(
+                name, item, [*path, str(index)], singular_name=True, parent=parent
+            )
+            for index, item in enumerate(target_items)
+        ]
+
     def parse_array_fields(
         self, name: str, obj: JsonSchemaObject, path: List[str]
     ) -> DataModelFieldBase:
-        def parse_field(index: int, item: JsonSchemaObject) -> DataType:
-            field_path = [*path, str(index)]
-            if item.has_constraint and (obj.has_constraint or self.field_constraints):
-                return self.parse_root_type(
-                    self.model_resolver.add(
-                        field_path, name, class_name=True, singular_name=True,
-                    ).name,
-                    item,
-                    field_path,
-                )
-            elif item.ref:
-                return self.get_ref_data_type(item.ref)
-            elif isinstance(item, JsonSchemaObject) and item.properties:
-                return self.parse_object(name, item, field_path, singular_name=True,)
-            elif item.anyOf:
-                return self.parse_any_of(name, item, field_path)
-            elif item.oneOf:
-                return self.parse_one_of(name, item, field_path)
-            elif item.allOf:
-                return self.parse_all_of(
-                    self.model_resolver.add(field_path, name, singular_name=True).name,
-                    item,
-                    field_path,
-                    ignore_duplicate_model=True,
-                )
-            elif item.enum:
-                if self.should_parse_enum_as_literal(item):
-                    return self.data_type(literals=item.enum)
-                else:
-                    return self.parse_enum(name, item, field_path, singular_name=True)
-            elif item.is_array:
-                return self.parse_array_fields(
-                    self.model_resolver.add(field_path, name, class_name=True).name,
-                    item,
-                    field_path,
-                ).data_type
-            else:
-                return self.get_data_type(item)
-
         if self.force_optional_for_required_fields:
             required: bool = False
             nullable: Optional[bool] = None
@@ -700,18 +628,36 @@ class JsonSchemaParser(Parser):
             else:
                 required = not obj.nullable and required
                 nullable = None
-        return self.data_model_field_type(
-            data_type=self.data_type(
-                data_types=[
-                    parse_field(index, item)
-                    for index, item in enumerate(
-                        [obj.items]
-                        if isinstance(obj.items, JsonSchemaObject)
-                        else obj.items or []
-                    )
-                ],
+
+        data_types: List[DataType] = [
+            self.data_type(
+                data_types=self.parse_list_item(
+                    name,
+                    [obj.items]
+                    if isinstance(obj.items, JsonSchemaObject)
+                    else obj.items or [],
+                    path,
+                    obj,
+                ),
                 is_list=True,
-            ),
+            )
+        ]
+        # TODO: decide special path word for a combined data model.
+        if obj.allOf:
+            data_types.append(
+                self.parse_all_of(name, obj, get_special_path('allOf', path))
+            )
+        elif obj.is_object:
+            data_types.append(
+                self.parse_object(name, obj, get_special_path('object', path))
+            )
+        if obj.enum:
+            data_types.append(
+                self.parse_enum(name, obj, get_special_path('enum', path))
+            )
+
+        return self.data_model_field_type(
+            data_type=self.data_type(data_types=data_types),
             example=obj.example,
             examples=obj.examples,
             default=obj.default,
@@ -724,10 +670,39 @@ class JsonSchemaParser(Parser):
         )
 
     def parse_array(
-        self, name: str, obj: JsonSchemaObject, path: List[str],
+        self,
+        name: str,
+        obj: JsonSchemaObject,
+        path: List[str],
+        original_name: Optional[str] = None,
     ) -> DataType:
-        field = self.parse_array_fields(name, obj, [*path, name])
-        reference = self.model_resolver.add(path, name, loaded=True)
+        reference = self.model_resolver.add(path, name, loaded=True, class_name=True)
+        field = self.parse_array_fields(original_name or name, obj, [*path, name])
+
+        if reference in [
+            d.reference for d in field.data_type.all_data_types if d.reference
+        ]:
+            # self-reference
+            field = self.data_model_field_type(
+                data_type=self.data_type(
+                    data_types=[
+                        self.data_type(
+                            data_types=field.data_type.data_types[1:], is_list=True
+                        ),
+                        *field.data_type.data_types[1:],
+                    ]
+                ),
+                example=field.example,
+                examples=field.examples,
+                default=field.default,
+                description=field.description,
+                title=field.title,
+                required=field.required,
+                constraints=field.constraints,
+                nullable=field.nullable,
+                strip_default_none=field.strip_default_none,
+            )
+
         data_model_root = self.data_model_root_type(
             reference=reference,
             fields=[field],
@@ -741,20 +716,33 @@ class JsonSchemaParser(Parser):
         return self.data_type(reference=reference)
 
     def parse_root_type(
-        self,
-        name: str,
-        obj: JsonSchemaObject,
-        path: List[str],
-        additional_properties: Optional[JsonSchemaObject] = None,
+        self, name: str, obj: JsonSchemaObject, path: List[str],
     ) -> DataType:
         if obj.ref:
             data_type: DataType = self.get_ref_data_type(obj.ref)
+        elif obj.is_object or obj.anyOf or obj.oneOf:
+            data_types: List[DataType] = []
+            object_path = [*path, name]
+            if obj.is_object:
+                data_types.append(
+                    self.parse_object(
+                        name, obj, get_special_path('object', object_path)
+                    )
+                )
+            if obj.anyOf:
+                data_types.extend(
+                    self.parse_any_of(name, obj, get_special_path('anyOf', object_path))
+                )
+            if obj.oneOf:
+                data_types.extend(
+                    self.parse_one_of(name, obj, get_special_path('oneOf', object_path))
+                )
+            if len(data_types) > 1:
+                data_type = self.data_type(data_types=data_types)
+            else:  # pragma: no cover
+                data_type = data_types[0]
         elif obj.type:
             data_type = self.get_data_type(obj)
-        elif obj.anyOf:
-            data_type = self.parse_any_of(name, obj, [*path, name])
-        elif obj.oneOf:
-            data_type = self.parse_one_of(name, obj, [*path, name])
         else:
             data_type = self.data_type_manager.get_data_type(Types.any)
         if self.force_optional_for_required_fields:
@@ -763,9 +751,9 @@ class JsonSchemaParser(Parser):
             required = not obj.nullable and not (
                 obj.has_default and self.apply_default_values_for_required_fields
             )
-        reference = self.model_resolver.add(path, name, loaded=True)
+        reference = self.model_resolver.add(path, name, loaded=True, class_name=True)
         self.set_title(name, obj)
-        self.set_additional_properties(name, additional_properties or obj)
+        self.set_additional_properties(name, obj)
         data_model_root_type = self.data_model_root_type(
             reference=reference,
             fields=[
@@ -851,25 +839,17 @@ class JsonSchemaParser(Parser):
                 )
             )
 
-        if not nullable:
-            reference = self.model_resolver.add(
-                path,
-                name,
-                class_name=True,
-                singular_name=singular_name,
-                singular_name_suffix='Enum',
-                loaded=True,
-            )
+        def create_enum(reference_: Reference) -> DataType:
             enum = Enum(
-                reference=reference,
+                reference=reference_,
                 fields=enum_fields,
                 path=self.current_source_path,
                 description=obj.description if self.use_schema_description else None,
             )
             self.results.append(enum)
-            return self.data_type(reference=reference)
+            return self.data_type(reference=reference_)
 
-        root_reference = self.model_resolver.add(
+        reference = self.model_resolver.add(
             path,
             name,
             class_name=True,
@@ -877,26 +857,24 @@ class JsonSchemaParser(Parser):
             singular_name_suffix='Enum',
             loaded=True,
         )
+
+        if not nullable:
+            return create_enum(reference)
+
         enum_reference = self.model_resolver.add(
             [*path, 'Enum'],
-            f'{root_reference.name}Enum',
+            f'{reference.name}Enum',
             class_name=True,
             singular_name=singular_name,
             singular_name_suffix='Enum',
             loaded=True,
         )
-        enum = Enum(
-            reference=enum_reference,
-            fields=enum_fields,
-            path=self.current_source_path,
-            description=obj.description if self.use_schema_description else None,
-        )
-        self.results.append(enum)
+
         data_model_root_type = self.data_model_root_type(
-            reference=root_reference,
+            reference=reference,
             fields=[
                 self.data_model_field_type(
-                    data_type=self.data_type(reference=enum_reference),
+                    data_type=create_enum(enum_reference),
                     description=obj.description,
                     example=obj.example,
                     examples=obj.examples,
@@ -912,7 +890,7 @@ class JsonSchemaParser(Parser):
             path=self.current_source_path,
         )
         self.results.append(data_model_root_type)
-        return self.data_type(reference=root_reference)
+        return self.data_type(reference=reference)
 
     def _get_ref_body(self, resolved_ref: str) -> Dict[Any, Any]:
         if is_url(resolved_ref):
@@ -950,17 +928,20 @@ class JsonSchemaParser(Parser):
                     if is_url(ref):
                         relative_path, object_path = ref.split('#')
                         relative_paths = [relative_path]
+                        base_path = None
                     else:
                         if self.model_resolver.is_external_root_ref(ref):
                             relative_path, object_path = ref[:-1], ''
                         else:
                             relative_path, object_path = ref.split('#')
                         relative_paths = relative_path.split('/')
-
-                    with self.model_resolver.base_url_context(relative_path):
+                        base_path = Path(*relative_paths).parent
+                    with self.model_resolver.current_base_path_context(
+                        base_path
+                    ), self.model_resolver.base_url_context(relative_path):
                         self._parse_file(
                             self._get_ref_body(relative_path),
-                            self.model_resolver.add_ref(obj.ref).name,
+                            self.model_resolver.add_ref(ref, resolved=True).name,
                             relative_paths,
                             object_path.split('/') if object_path else None,
                         )
@@ -1015,15 +996,16 @@ class JsonSchemaParser(Parser):
         self.parse_obj(name, JsonSchemaObject.parse_obj(raw), path)
 
     def parse_obj(self, name: str, obj: JsonSchemaObject, path: List[str],) -> None:
-        name = self.model_resolver.add(path, name, class_name=True).name
-        if obj.is_object:
-            self.parse_object(name, obj, path)
-        elif obj.is_array:
+        if obj.is_array:
             self.parse_array(name, obj, path)
-        elif obj.enum:
-            self.parse_enum(name, obj, path)
         elif obj.allOf:
             self.parse_all_of(name, obj, path)
+        elif obj.oneOf:
+            self.parse_root_type(name, obj, path)
+        elif obj.is_object:
+            self.parse_object(name, obj, path)
+        elif obj.enum:
+            self.parse_enum(name, obj, path)
         else:
             self.parse_root_type(name, obj, path)
         self.parse_ref(obj, path)
@@ -1034,7 +1016,8 @@ class JsonSchemaParser(Parser):
         ):
             self.current_source_path = Path()
             self.model_resolver.after_load_files = {
-                s.path.as_posix() for s in self.iter_source
+                self.base_path.joinpath(s.path).resolve().as_posix()
+                for s in self.iter_source
             }
 
         for source in self.iter_source:
@@ -1044,21 +1027,22 @@ class JsonSchemaParser(Parser):
                 path_parts = list(source.path.parts)
             if self.current_source_path is not None:
                 self.current_source_path = source.path
-            with self.model_resolver.current_root_context(path_parts):
+            with self.model_resolver.current_base_path_context(
+                source.path.parent
+            ), self.model_resolver.current_root_context(path_parts):
                 self.raw_obj = load_yaml(source.text)
-                if self.class_name:
-                    obj_name = self.class_name
-                elif self.custom_class_name_generator:
-                    obj_name = self.custom_class_name_generator(
-                        self.raw_obj.get('title', 'Model')
-                    )
-                else:
-                    # backward compatible
+                if self.custom_class_name_generator:
                     obj_name = self.raw_obj.get('title', 'Model')
+                else:
+                    if self.class_name:
+                        obj_name = self.class_name
+                    else:
+                        # backward compatible
+                        obj_name = self.raw_obj.get('title', 'Model')
+                        if not self.model_resolver.validate_name(obj_name):
+                            obj_name = title_to_class_name(obj_name)
                     if not self.model_resolver.validate_name(obj_name):
-                        obj_name = title_to_class_name(obj_name)
-                if not self.model_resolver.validate_name(obj_name):
-                    raise InvalidClassNameError(obj_name)
+                        raise InvalidClassNameError(obj_name)
                 self._parse_file(self.raw_obj, obj_name, path_parts)
 
         self._resolve_unparsed_json_pointer()
@@ -1073,7 +1057,9 @@ class JsonSchemaParser(Parser):
             if self.current_source_path is not None:
                 self.current_source_path = source.path
 
-            with self.model_resolver.current_root_context(path_parts):
+            with self.model_resolver.current_base_path_context(
+                source.path.parent
+            ), self.model_resolver.current_root_context(path_parts):
                 for reserved_ref in sorted(reserved_refs):
                     if self.model_resolver.add_ref(reserved_ref, resolved=True).loaded:
                         continue
@@ -1112,7 +1098,9 @@ class JsonSchemaParser(Parser):
         else:
             path = path_parts
         with self.model_resolver.current_root_context(path_parts):
-            obj_name = self.model_resolver.add(path, obj_name, unique=False).name
+            obj_name = self.model_resolver.add(
+                path, obj_name, unique=False, class_name=True
+            ).name
             with self.root_id_context(raw):
 
                 # parse $id before parsing $ref
